from typing import List

from blue_options.terminal import show_usage, xtra

from roofai.semseg import Profile

list_of_profiles = [profile.name for profile in Profile]

device_and_profile_details = {
    "device: cpu | cuda": [],
    "profile: {}".format(" | ".join(list_of_profiles)): [],
}


def predict_options(
    mono: bool,
    cascade: bool = False,
):
    return "".join(
        [
            xtra(
                "device=<device>,{}profile=<profile>,".format(
                    "~download,dryrun," if not cascade else ""
                ),
                mono=mono,
            ),
            "upload",
        ]
    )


def help_predict(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "roofai",
            "semseg",
            "predict",
            f"[{predict_options(mono=mono)}]",
            "[..|<model-object-name>]",
            "[.|<dataset-object-name>]",
            "[-|<prediction-object-name>]",
        ],
        "semseg[<model-object-name>].predict(<dataset-object-name>) -> <prediction-object-name>.",
        device_and_profile_details,
        mono=mono,
    )


def train_options(
    mono: bool,
    show_download: bool = True,
):
    return "".join(
        [
            xtra(
                "device=<device>,{}dryrun,profile=<profile>,".format(
                    "~download," if show_download else ""
                ),
                mono=mono,
            ),
            "upload",
        ]
    )


def help_train(
    tokens: List[str],
    mono: bool,
) -> str:
    args = [
        "[--activation <sigmoid>]",
        "[--classes <one+two+three+four>]",
        "[--encoder_name <se_resnext50_32x4d>]",
        "[--encoder_weights <imagenet>]",
        "[--epoch_count <-1>]",
    ]

    return show_usage(
        [
            "roofai",
            "semseg",
            "train",
            f"[{train_options(mono=mono)}]",
            "[.|<dataset-object-name>]",
            "[-|<model-object-name>]",
        ]
        + args,
        "semseg.train(<dataset-object-name>) -> <model-object-name>.",
        device_and_profile_details,
        mono=mono,
    )


help_functions = {
    "predict": help_predict,
    "train": help_train,
}



from typing import Dict, Tuple
import requests
import pprint

from blueness import module
from blue_objects.metadata import post_to_object

from roofai import NAME
from roofai.env import GOOGLE_MAPS_API_KEY
from roofai.logger import logger

NAME = module.name(__file__, NAME)

# https://developers.google.com/maps/documentation/geocoding/start
base_url = "https://maps.googleapis.com/maps/api/geocode/json?"


def geocode(
    address: str,
    object_name: str = "",
    verbose: bool = False,
) -> Tuple[bool, float, float, Dict]:
    success = True

    params = {
        "address": address,
        "key": GOOGLE_MAPS_API_KEY,
    }
    logger.info(f"{NAME}.geocode({address})")

    response = requests.get(base_url, params=params)

    description_of_failure: str = ""
    metadata: Dict = {}
    lat: float = 0.0
    lon: float = 0.0
    if response.status_code != 200:
        success = False
        description_of_failure = "status_code={}, text={}".format(
            response.status_code,
            response.text,
        )

    if success:
        metadata = response.json()
        if verbose:
            pprint.pprint(metadata)

        if metadata["status"] != "OK":
            success = False
            description_of_failure = "status={}".format(
                metadata["status"],
            )

    if success:
        if len(metadata["results"]) == 0:
            success = False
            description_of_failure = "no results found"
        elif len(metadata["results"]) > 1:
            logger.warning(
                "{} result(s), using the first one.".format(len(metadata["results"]))
            )

    if success:
        try:
            lat = metadata["results"][0]["geometry"]["location"]["lat"]
            lon = metadata["results"][0]["geometry"]["location"]["lng"]
        except Exception as e:
            success = False
            description_of_failure = str(e)

    if not success:
        logger.error(
            "{}.geocode({}) failed: {}".format(
                NAME,
                address,
                description_of_failure,
            )
        )
        return False, 0, 0, {"error": description_of_failure}

    logger.info(
        "{}.geocode({}): lat:{:.6f}, lon:{:.6f}".format(
            NAME,
            address,
            lat,
            lon,
        )
    )

    if object_name:
        success = post_to_object(
            object_name,
            "geocoding",
            {
                "address": address,
                "lat": lat,
                "lon": lon,
                "metadata": metadata,
            },
        )

    return success, lat, lon, metadata



import pytest

from blue_objects import objects

from roofai import env
from roofai.tests.caching import cache_pretrainedmodels
from roofai.semseg.interface import predict, train
from roofai.semseg.model import SemSegModel


@pytest.mark.parametrize(
    "dataset_object_name, classes",
    [
        (
            # roofai dataset ingest \
            #  source=palisades-dataset-v1,upload \
            #  palisades-dataset-v1-1000X-test \
            #  --test_count 100 \
            #  --train_count 800 \
            #  --val_count 100
            "palisades-dataset-v1-1000X-test",
            ["affected"],
        ),
        (
            env.TEST_roofAI_ingest_AIRS_v1,
            ["roof"],
        ),
        (
            env.TEST_roofAI_ingest_CamVid_v1,
            ["car"],
        ),
    ],
)
def test_semseg_train(dataset_object_name, classes):
    assert cache_pretrainedmodels()

    assert objects.download(dataset_object_name)

    model_object_name = objects.unique_object("test_semseg_train-model")

    model = train(
        dataset_path=objects.object_path(dataset_object_name),
        model_path=objects.object_path(model_object_name),
        classes=classes,
    )
    assert isinstance(model, SemSegModel)

    predict(
        model_path=objects.object_path(model_object_name),
        dataset_path=objects.object_path(dataset_object_name),
        prediction_path=objects.object_path(objects.unique_object()),
        device="cpu",
    )



"""
Data augmentation is a powerful technique to increase the amount of your data and prevent model overfitting.  
If you not familiar with such trick read some of these articles:
 - [The Effectiveness of Data Augmentation in Image Classification using Deep
Learning](http://cs231n.stanford.edu/reports/2017/pdfs/300.pdf)
 - [Data Augmentation | How to use Deep Learning when you have Limited Data](https://medium.com/nanonets/how-to-use-deep-learning-when-you-have-limited-data-part-2-data-augmentation-c26971dc8ced)
 - [Data Augmentation Experimentation](https://towardsdatascience.com/data-augmentation-experimentation-3e274504f04b)

Since our dataset is very small we will apply a large number of different augmentations:
 - horizontal flip
 - affine transforms
 - perspective transforms
 - brightness/contrast/colors manipulations
 - image bluring and sharpening
 - gaussian noise
 - random crops

All this transforms can be easily applied with [**Albumentations**](https://github.com/albu/albumentations/) - fast augmentation library.
For detailed explanation of image transformations you can look at [kaggle salt segmentation exmaple](https://github.com/albu/albumentations/blob/master/notebooks/example_kaggle_salt.ipynb) provided by [**Albumentations**](https://github.com/albu/albumentations/) authors.

Copied with minor modification from ../../notebooks/semseg.ipynb
"""

import albumentations as albu


def get_training_augmentation():
    train_transform = [
        albu.HorizontalFlip(p=0.5),
        albu.ShiftScaleRotate(
            scale_limit=0.5, rotate_limit=0, shift_limit=0.1, p=1, border_mode=0
        ),
        albu.PadIfNeeded(
            min_height=320, min_width=320, always_apply=True, border_mode=0
        ),
        albu.RandomCrop(height=320, width=320, always_apply=True),
        albu.GaussNoise(p=0.2),
        albu.Perspective(p=0.5),
        albu.OneOf(
            [
                albu.CLAHE(p=1),
                albu.RandomBrightnessContrast(p=1),
                albu.RandomGamma(p=1),
            ],
            p=0.9,
        ),
        albu.OneOf(
            [
                albu.Sharpen(p=1),
                albu.Blur(blur_limit=3, p=1),
                albu.MotionBlur(blur_limit=3, p=1),
            ],
            p=0.9,
        ),
        albu.OneOf(
            [
                albu.RandomBrightnessContrast(p=1),
                albu.HueSaturationValue(p=1),
            ],
            p=0.9,
        ),
    ]
    return albu.Compose(train_transform)


def get_validation_augmentation():
    """Add paddings to make image shape divisible by 32"""
    test_transform = [albu.PadIfNeeded(384, 480)]
    return albu.Compose(test_transform)


def to_tensor(x, **kwargs):
    return x.transpose(2, 0, 1).astype("float32")


def get_preprocessing(preprocessing_fn):
    """Construct preprocessing transform

    Args:
        preprocessing_fn (callbale): data normalization function
            (can be specific for each pretrained neural network)
    Return:
        transform: albumentations.Compose

    """

    _transform = [
        albu.Lambda(image=preprocessing_fn),
        albu.Lambda(image=to_tensor, mask=to_tensor),
    ]
    return albu.Compose(_transform)



import os.path
from typing import List
from roofai.semseg.model import SemSegModel
from roofai.semseg.train import SemSegModelTrainer
from roofai.semseg import Profile
from typing import Tuple, Any


def predict(
    model_path: str,
    dataset_path: str,
    prediction_path: str,
    device: str,
    profile: Profile = Profile.VALIDATION,
    in_notebook: bool = False,
):
    model = SemSegModel(
        model_filename=os.path.join(model_path, "model.pth"),
        profile=profile,
        device=device,
    )

    model.predict(
        dataset_path=dataset_path,
        output_path=prediction_path,
        in_notebook=in_notebook,
    )


def train(
    dataset_path: str,
    model_path: str,
    classes: List[str],
    encoder_name="se_resnext50_32x4d",
    encoder_weights="imagenet",
    activation="sigmoid",  # could be None for logits or 'softmax2d' for multi-class segmentation
    device="cpu",  # 'cuda'
    profile: Profile = Profile.VALIDATION,
    in_notebook: bool = False,
    epoch_count: int = -1,
) -> SemSegModel:
    trainer = SemSegModelTrainer(
        dataset_path=dataset_path,
        model_path=model_path,
        classes=classes,
        in_notebook=in_notebook,
        profile=profile,
    )

    model = trainer.train(
        classes=classes,
        encoder_name=encoder_name,
        encoder_weights=encoder_weights,
        activation=activation,
        device=device,
        in_notebook=in_notebook,
        epoch_count=epoch_count,
    )

    model.predict(
        dataset_path=dataset_path,
        output_path=model_path,
        in_notebook=in_notebook,
    )

    return model



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit
from blue_objects import objects

from roofai import NAME
from roofai.google_maps.api.geocoding import geocode
from roofai.google_maps.api.static import get as get_static_image
from roofai.google_maps.semseg.dataset import ingest_dataset
from roofai.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="get_static_image | geocode | ingest_dataset",
)
parser.add_argument(
    "--lat",
    type=float,
    default=0,
)
parser.add_argument(
    "--lon",
    type=float,
    default=0,
)
parser.add_argument(
    "--filename",
    type=str,
    default="",
)
parser.add_argument(
    "--object_name",
    type=str,
    default="",
)
parser.add_argument(
    "--zoom",
    type=int,
    default=20,
)
parser.add_argument(
    "--maptype",
    type=str,
    default="satellite",
)
parser.add_argument(
    "--size",
    type=str,
    default="640x640",
)
parser.add_argument(
    "--address",
    type=str,
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--count",
    type=int,
    default=10,
)
args = parser.parse_args()

success = False
if args.task == "get_static_image":
    success, _, _ = get_static_image(
        lat=args.lat,
        lon=args.lon,
        filename=(
            objects.path_of(
                object_name=args.object_name,
                filename=args.filename,
            )
            if args.filename
            else ""
        ),
        zoom=args.zoom,
        maptype=args.maptype,
        size=args.size,
    )
elif args.task == "geocode":
    success, _, _, _ = geocode(
        address=args.address,
        object_name=args.object_name,
        verbose=args.verbose == 1,
    )
elif args.task == "ingest_dataset":
    success = ingest_dataset(
        object_name=args.object_name,
        lat=args.lat,
        lon=args.lon,
        count=args.count,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Union, Dict
import requests
from PIL import Image
import numpy as np
import io
import math

from blueness import module
from blue_options import string

from roofai import NAME
from roofai.env import GOOGLE_MAPS_API_KEY
from roofai.logger import logger

NAME = module.name(__file__, NAME)

base_url = "https://maps.googleapis.com/maps/api/staticmap?"


# https://developers.google.com/maps/documentation/maps-static/start
def get(
    lat: float,
    lon: float,
    filename: str = "",
    zoom: int = 20,
    maptype: str = "satellite",
    size: str = "640x640",
) -> Union[bool, np.ndarray, Dict]:
    image: np.ndarray = np.array(())

    params = {
        "center": f"{lat},{lon}",
        "zoom": zoom,
        "size": size,
        "maptype": maptype,
        "key": GOOGLE_MAPS_API_KEY,
    }
    logger.info(
        "{}.get: {}".format(
            NAME,
            ", ".join(
                [f"{key}:{value}" for key, value in params.items() if key != "key"]
            ),
        )
    )

    response = requests.get(base_url, params=params)

    if response.status_code != 200:
        logger.error(
            "status_code={}, text={}".format(
                response.status_code,
                response.text,
            )
        )
        return False, image, {}

    try:
        image = np.array(Image.open(io.BytesIO(response.content)))
    except Exception as e:
        logger.error(e)
        return False, image, {}
    logger.info(string.pretty_shape_of_matrix(image))

    if filename:
        try:
            with open(filename, "wb") as file:
                file.write(response.content)
        except Exception as e:
            logger.error(e)
            return False, image, {}

        logger.info(f"-> {filename}")

    # https://groups.google.com/g/google-maps-js-api-v3/c/hDRO4oHVSeM?pli=1
    gsd = 156543.03392 * math.cos(lat * np.pi / 180) / (2**zoom)
    logger.info(f"gsd: {gsd:.2f} m")

    return True, image, {"gsd": gsd}



from typing import List

from blue_options.terminal import show_usage, xtra


def help_geocode(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("-", mono=mono)

    args = [
        "[--address <address>]",
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@gmaps",
            "geocode",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + args,
        "geocode <address>.",
        mono=mono,
    )



import pytest
import numpy as np

from blue_options import string
from blue_objects import objects

from roofai import env
from roofai.google_maps.api.geocoding import geocode


@pytest.mark.parametrize(
    ["address", "expected_success"],
    [
        ["350 W Georgia St, Vancouver, BC V6B 6B1, Canada", True],
        ["void", False],
    ],
)
def test_google_maps_geocode(
    address: str,
    expected_success: bool,
):
    for object_name in [
        "",
        objects.unique_object("test_google_maps_geocode"),
    ]:
        success, lat, lon, _ = geocode(
            address=address,
            object_name=object_name,
            verbose=True,
        )
        assert success == expected_success

        if expected_success:
            assert isinstance(lat, float)
            assert isinstance(lon, float)



from typing import List

from blue_options.terminal import show_usage, xtra


def help_ingest_AIRS_or_distributed(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("download,dryrun,", mono=mono),
            "source=<source>",
            xtra(",target=<target>,upload", mono=mono),
        ]
    )

    args = [
        "[--test_count <10>]",
        "[--train_count <80>]",
        "[--val_count <10>]",
    ]

    return show_usage(
        [
            "roofai",
            "dataset",
            "ingest",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + args,
        "ingest <source> -> <object-name>.",
        {
            "source: AIRS | <query-object-name>": [],
            "target: sagemaker | torch": [],
        },
        mono=mono,
    )


def help_ingest_CamVid(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("dryrun,", mono=mono),
            "source=CamVid",
            xtra(",upload", mono=mono),
        ]
    )

    return show_usage(
        [
            "roofai",
            "dataset",
            "ingest",
            f"[{options}]",
            "[-|<object-name>]",
        ],
        "ingest CamVid -> <object-name>.",
        mono=mono,
    )


def help_ingest_gmaps(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("dryrun,", mono=mono),
            "source=gmaps",
            xtra(",upload", mono=mono),
        ]
    )

    ingest_options = "count=<count>,lat=<lat>,lon=<lon>"

    return show_usage(
        [
            "roofai",
            "dataset",
            "ingest",
            f"[{options}]",
            "[-|<object-name>]",
            f"[{ingest_options}]",
        ],
        "ingest gmaps -> <object-name>.",
        mono=mono,
    )


def help_ingest(
    tokens: List[str],
    mono: bool,
) -> str:
    return "\n".join(
        [
            help_ingest_AIRS_or_distributed(tokens, mono),
            help_ingest_CamVid(tokens, mono),
            help_ingest_gmaps(tokens, mono),
        ]
    )


def help_review(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            "download",
            xtra(",dryrun", mono=mono),
        ]
    )

    args = [
        "[--count <1>]",
        "[--index <index>]",
        "[--subset <subset>]",
    ]

    return show_usage(
        [
            "roofai",
            "dataset",
            "review",
            f"[{options}]",
            "[.|<object-name>]",
        ]
        + args,
        "review <object-name>.",
        mono=mono,
    )


help_functions = {
    "ingest": help_ingest,
    "review": help_review,
}



from typing import List

from blue_options.terminal import show_usage, xtra
from abcli.help.generic import help_functions as generic_help_functions

from roofai import ALIAS
from roofai.help.dataset import help_functions as help_dataset
from roofai.help.google_maps import help_functions as help_google_maps
from roofai.help.semseg import help_functions as help_semseg


help_functions = generic_help_functions(plugin_name=ALIAS)

help_functions.update(
    {
        "dataset": help_dataset,
        "google_maps": help_google_maps,
        "semseg": help_semseg,
    }
)



from blueness import module

from roofai import NAME
from roofai.env import GOOGLE_MAPS_API_KEY
from roofai.logger import logger

NAME = module.name(__file__, NAME)


def ingest_dataset(
    lat: float,
    lon: float,
    count: int,
    object_name: str,
) -> bool:
    logger.info(
        "{}.ingest_dataset(lat:{:.6f}, lon:{:.6f}, count:{}) -> {}".format(
            NAME,
            lat,
            lon,
            count,
            object_name,
        )
    )

    logger.info("🪄")

    return True



from blue_objects import objects

from roofai.tests.caching import cache_pretrainedmodels
from roofai.semseg.interface import predict
from roofai import env


def test_semseg_predict():
    assert cache_pretrainedmodels()

    model_object_name = env.TEST_roofAI_semseg_model_CamVid_v1
    assert objects.download(model_object_name)

    dataset_object_name = env.TEST_roofAI_ingest_CamVid_v1
    assert objects.download(dataset_object_name)

    prediction_object_name = objects.unique_object("test_semseg_predict")

    predict(
        model_path=objects.object_path(model_object_name),
        dataset_path=objects.object_path(dataset_object_name),
        prediction_path=objects.object_path(prediction_object_name),
        device="cpu",
    )
